/* Changes made by TW for porting to Parallel computers

   add Wssmat_init() in An_circuit;
   add Ssmat_mv_2be() in met_mod;
   add Wsmat_put() in met_mod;

*/
extern Wssmat_init(/* int prec, loss, size, int *np */);
/*************************************************************/
/* Perform the em analysis.                                  */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr25may87 */ 

#include "local.h"
#include <math.h>
#include <stdio.h>  /*for NULL definition*/
#include "em.typ"
#include "son.h"
#include "hl.h"
#include "hardlock.h"

#ifdef CM5
#include "cm/CMutil.h"
#endif

/* MKS units used at all times */
#include "hharray.h"

static int cycles; /*For Zhat summation*/
static Mat_ptr Znhat[3],Zhat;  /*Required functions of Zhat*/
static Mat_ptr Lay[2],Kckz;  /*Layer to Layer coupling modification for Zhat*/
static Vec_ptr Layz1,Layz2,kcz_src,kcz_fld;

Rtyp min_lambda();
void loadhh(),loadh();

An_circuit( Box,cir,An,stime,F )  Box_info *Box;  Cir *cir;
				  An_info *An;  sTime *stime;  Rtyp F; /*MHz*/
/*************************************************************/
/* Perform the em analysis.                                  */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr25may87 */ 
/* jcr15jun88 -- Signficantly modified for multi-layer for Kzh -> j inf. */
/* jcr15feb90 -- Signficantly modified for vias */
/* jcr18jun90 -- S converted to rtyp (single precision)*/
{ static Ssmat_ptr S = NULL;  /*Single/double precision symmetric matrix ptr*/
  Mat_ptr ans,result;
  int Num_modes,Num_lev,nrow,ncol;
  Mat_ptr merge_ports();
  Vec_ptr kcz_src,kcz_fld;
  int i,Lsrc,Lfld;
  int j,quad_prec_sto;
  int start_time;
  msub_list *src,*fld;
  Rtyp rm,rn;
  int hldata;
  int np;

  if (An->cans != NULL)
  { announce_bug("An_circuit(0)"); }

  Znhat[0] = Znhat[1] = Znhat[2] = NULL;
  Lay[0] = Lay[1] = NULL;
  Kckz = NULL;
  Layz1 = Layz2  = kcz_src = kcz_fld = NULL;

  /*Number of met levels doesn't include gnd via level*/
  Num_lev = Box->Num_levels;

  /*Check for a zero frequency analysis*/
  quad_prec_sto = Box->quad_prec;
  if ( F < Box->qfmin )
    Box->quad_prec = 1;

  /*Set Zhat interpolation parameters*/
  cycles = An->Cyc;  /*Global*/
  rm = Box->maxm = cycles * Box->Capm;  rn = Box->maxn = cycles * Box->Capn;
  rm /= Box->A;  rm *= rm;  rn /= Box->B;  rn *= rn;
  Box->kc_max = PI * sqrt( rm + rn );

  /*Set maxm,maxn so that they are at least twice cutoff*/
  i = cycles * (int)(8.0 * Box->A / min_lambda(Box->Eps,Box->Mu,Num_lev,F));
  if ( i < 8*cycles ) i = 8 * cycles;
  if ( i < Box->maxm ) Box->maxm = i;
  i = cycles * (int)(8.0 * Box->B / min_lambda(Box->Eps,Box->Mu,Num_lev,F));
  if ( i < 8*cycles ) i = 8 * cycles;
  if ( i < Box->maxn ) Box->maxn = i;

/*i = 18;
Box->maxn = Box->maxm = i;*/

  /* DOS SECURITY */
#ifdef DOS
  {
    int data;
    int port;
    int hladdress = HL_ADDRESS;
    int registr;
    int tmpdatastr[9],status,blkcnt;
    port = 0;
    blkcnt = 1;
    strcpy(tmpdatastr,"12345678");
    hlcrypt (&status,&port,&hladdress,tmpdatastr,&blkcnt);
      
    if ( status == 0 ) {
      intoff();
      hlon(&port,&hladdress);
      registr = HL_LOC_8232;
      hlrwd(&port,&registr,&data);
      hloff(&port);
      inton();
    } else {
      fprintf(stderr, "ERROR: Hardware key not found.\n");
      fprintf(stderr, "Em is unexpectly terminated.  You must keep the hardware key \n");
      fprintf(stderr, "attached to the computer during the entire em analysis.\n");
      exit_program(1);
    }
    hldata = data;
  }
#endif

  /*Kill interpolation if cycles = 5*/
  if ( cycles >= 5 )
  { Box->maxm = cycles * Box->Capm;  Box->maxn = cycles * Box->Capn; }

  rm = Box->maxm * PI / Box->A;  rn = Box->maxn * PI / Box->B;
  Box->kc_min = (rm<rn) ? rm : rn;
  Box->num_zterp = 0;

  if (Box->maxm < Box->Capm || Box->maxn < Box->Capn)
  { i = 2 * cycles * Box->Capn - Box->maxn;
    Box->num_zterp = 2 * cycles * Box->Capm - Box->maxm;
    if (i>Box->num_zterp) Box->num_zterp = i;
    /*Extend lower and upper limits so only center of 4 pt interp needed*/
    rm = 1.1 * (1.0/Box->kc_max - 1.0/Box->kc_min)/( (Rtyp)(Box->num_zterp-1) );
    Box->kc_min = 1.0 / ( 1.0/Box->kc_min - rm );
    Box->kc_max = 1.0 / ( 1.0/Box->kc_max + rm );
    Box->num_zterp += 2;
  }
  if ( Box->num_zterp <= 3 )
  { Box->num_zterp = 0;
    Box->maxm = cycles * Box->Capm;
    Box->maxn = cycles * Box->Capn;
  }

  f_ck_on();  /*Turn on Vec floating pt error cking/correcting*/

  if (An->State) printf("Analyzing %10lf %cHz.\n",(An->Funit=='M') ? F : F/1000,An->Funit);
  Num_modes = Box->maxm;
  Num_modes *= (Box->Xsym) ? (Box->maxn >> 1) : Box->maxn;
  Num_modes += Box->num_zterp; 
  Znhat[0] = Mat_open(Num_lev+2,Num_modes);  /*Num_lev includes top and bottom gnd planes*/
  Znhat[1] = Mat_open(Num_lev+2,Num_modes);
  Znhat[2] = Mat_open(Num_lev+1,Num_modes);  /*For via-via coupling on same layer only*/
  Lay[0] = Mat_open(Num_lev+1,Num_modes);
  Lay[1] = Mat_open(Num_lev+1,Num_modes);
  Kckz = Mat_open(Num_lev+1,Num_modes);
  Load_znhat(Box,cir,F,Znhat,Lay,Kckz);

  /*Do pirate ck*/
  if (!doneck()) for (Lsrc=0; Lsrc<Num_lev+2; Lsrc++)
  { Vec_smult(Box->A,Box->B,Znhat[0]->Row[Lsrc],Znhat[0]->Row[Lsrc]);
    Vec_smult(Box->B,*(Box->H),Znhat[1]->Row[Lsrc],Znhat[1]->Row[Lsrc]);
  }

  f_ck_off();  /*Turn off Vec floating pt error cking/correcting*/

#ifdef EMPAR
#ifdef EM_MASPAR
  callRequest(Wssmat_init,16,!(An->Mem_saver),!Box->Lossless,Box->mat_size,&np);
#endif /* EM_MASPAR */
#endif /* EMPAR */

  /*Use double precision only if not Mem_saver*/
  if ( S == NULL )
  { S = (Ssmat_ptr)cnew(1,sizeof(Ssmat_struct));
    Ssmat_init( S,!(An->Mem_saver) );
  }

  Ssmat_getmem(S,Box->mat_size,(Box->met_loss || !Box->Lossless),1);

  /* DOS SECURITY */
#ifdef DOS
  {
    char *msgstr;
    int ind;

    /* If hldata is worng print out "Invalid hardware key". Add 1 to	*/
    /* each letter in error message to make it more difficult to find	*/
    /* by a hacker							*/

    /* hldata = 8232 */
    if ( ((hldata / 7 - 3)-212) % 135 == 29 ) {
      msgstr = string("Jowbmje ibsexbsf lfz");
      for ( ind=0; ind<strlen(msgstr); ++ind ) 
      { if ( msgstr[ind] != ' ' ) 
	  --msgstr[ind];
      }
      fprintf(stderr,"%s.\n",msgstr);
      hldata = 1;
    }
  }
#endif

  /*Re part might have been allocated during a previous frequency for
   *metal loss, if so, it needs to be zeroed out here because do_coup
   *won't overwrite it if Box->Lossless.
   */
#ifdef EMPAR
  if ( Box->Lossless && S->Re )
    for (i=0; i<((S->len < S->nmem)?S->len:S->nmem); i++) *(S->Re+i) = 0;
  if ( Box->Lossless && S->re )
    for (i=0; i<((S->len < S->nmem)?S->len:S->nmem); i++) *(S->re+i) = 0;
#else /* NOT EMPAR */
  if ( Box->Lossless && S->Re )
    for (i=0; i<S->len; i++) *(S->Re+i) = 0;
  if ( Box->Lossless && S->re )
    for (i=0; i<S->len; i++) *(S->re+i) = 0;
#endif

  Layz1 = Vec_open(Num_modes);
  Layz2 = Vec_open(Num_modes);
  Zhat = Mat_open(5,Num_modes);

  init_do_coup(0,0,cir->max_wt_len,cir->wt0);

  if (stime)
  { stime = GetTime( stime );
    An->mode_calc_time = stime->user;
    if (An->State && An->mode_calc_time > 0)
      printf("    Waveguide mode calculation time = %d Seconds.\n",An->mode_calc_time);
    start_time = stime->user;  /*To figure out next delta*/
  }

  /* DOS SECURITY */
#ifdef DOS
  if ( hldata == 1 ) exit_program(1);
#endif /* DOS */

  /*Lev Num_lev is for gnd vias only*/
  for (Lsrc=0,src=cir->lev0; Lsrc<=Num_lev; Lsrc++,src++)
  for (Lfld=Lsrc,fld=src; Lfld>=0; Lfld--,fld--)
  { if (Lsrc==Lfld)
    { f_ck_on();  /*Turn on Vec floating pt error cking/correcting*/
      Vec_move(Znhat[0]->Row[Lsrc+1],Zhat->Row[0]);  /*Zhat_te*/
      Vec_move(Znhat[1]->Row[Lsrc+1],Zhat->Row[1]);  /*Zhat_tm*/
      Vec_mult(Zhat->Row[1],Lay[1]->Row[Lsrc],Layz1);
      if (cir->vias->n)
      { Vec_sub(Layz1,Zhat->Row[1],Zhat->Row[2]);  /*Zhat2 complete*/
        Vec_add(Layz1,Zhat->Row[2],Zhat->Row[3]);
        Vec_move(Znhat[1]->Row[Lsrc],Layz2);
        Vec_sub(Zhat->Row[3],Layz2,Zhat->Row[3]);  /*Zhat3 complete*/
        Vec_move(Znhat[2]->Row[Lsrc],Zhat->Row[4]); /*Used when Lsrc==Lfld*/
      }
      f_ck_off();  /*Turn off Vec floating pt error cking/correcting*/
    }
    else /*Modify for coupling to next level*/
    { f_ck_on();  /*Turn on Vec floating pt error cking/correcting*/
      Vec_mult(Zhat->Row[0],Lay[0]->Row[Lfld+1],Zhat->Row[0]); /*Zte_hat done*/
      Vec_move(Layz1,Zhat->Row[1]);  /*Ztm_hat done*/
      if (cir->vias->n)
      { /*Set up Jz to Vx,Vy in Zhat->Row[4], used only when Lsrc!=Lfld*/
        if ( Lsrc==(Lfld+1) ) Vec_sub(Layz2,Layz1,Zhat->Row[4]);
        else                  Vec_mult(Zhat->Row[4],Lay[1]->Row[Lfld+1],Zhat->Row[4]);
        Vec_mult(Layz1,Lay[1]->Row[Lfld],Layz1);
        Vec_sub(Layz1,Zhat->Row[1],Zhat->Row[2]);  /*Zhat2 complete*/
        Vec_add(Zhat->Row[2],Layz2,Zhat->Row[3]);
        Vec_mult(Layz2,Lay[1]->Row[Lfld],Layz2);
        Vec_sub(Zhat->Row[3],Layz2,Zhat->Row[3]);  /*Zhat3 complete*/
      }
      else
        Vec_mult(Layz1,Lay[1]->Row[Lfld],Layz1);
      f_ck_off();  /*Turn off Vec floating pt error cking/correcting*/
    }

    if ( (src->n > 0) && ( fld->n > 0) )
    { if (An->State) printf("  Loading matrix, source level %d, field level %d.\n",Lsrc,Lfld);
      kcz_src = Kckz->Row[Lsrc];
      kcz_fld = Kckz->Row[Lfld];
      if (Box->quad_prec) qloads( src,fld,Box,kcz_src,kcz_fld,S );
      else                 loads( src,fld,Box,kcz_src,kcz_fld,S );
    }

  }

  if ( stime )
  { stime = GetTime( stime );
    An->mat_fill_time = stime->user - start_time;
    start_time = stime->user;
    if ( An->State && An->mat_fill_time > 0 )
      printf("    Matrix fill time = %d Seconds.\n",An->mat_fill_time);
  }

  Mat_close(Znhat[0]);  Mat_close(Znhat[1]);  Mat_close(Znhat[2]);
  Mat_close(Zhat);
  Mat_close(Lay[0]);  Mat_close(Lay[1]);
  Vec_close(Layz1);  Vec_close(Layz2);
  Mat_close(Kckz);

  /* Layer/Level numbering for data in above loop (for 3 dielectrics, Num_lev=2).
   *
   *   Znhat[.]->Row[x]   Lsrc/Lfld   Lay[.]->Row[y]
   *          0             n/a      ----------------   Top Cover
   *          1              0       ----------------   <-- First msubs
   *                                        1
   *          2              1       ----------------
   *                                        3
   *          3              2       ----------------   Bottom Cover (+ Gnd vias)
   */

  /*Take into account metalization loss*/
  if (Box->met_loss || Box->met_react) met_mod(cir,Box,F,S);

#ifdef CM5
  CM_mv_2_be( S, 1 );     /* Flush the buffer! 1 means done_mapping */
#endif

/* Ssmat_output(S); */

  /*Solve and return*/
  if (An->State) printf("  Reducing matrix.\n");
  /*Evaluate the number of port subsections*/
  for (i=0,nrow=0; i<cir->ports->n; i++) nrow += (cir->ports->port0+i)->nsub;
  /*If An->jans is already allocated, we don't want to solve for jxy, because
    this is then an analysis of a de-embedding standard*/
  ncol = (An->jxy && An->jans==NULL) ? S->n : nrow;

  /* Added by src 2/10/94 for debugging...*/
/* #ifdef CM5_DEBUG */
/*  CMSsmat_output(S); */
/*   CMSsmat_check_triangles(S);
  exit(0);
#else
  Ssmat_output(S);
#endif */
  /* end src debugging statements */

  result = Ssmat_lud(S,nrow,ncol);

/*mat_output(result);*/

  ans = NULL;

  An->cans = merge_ports(result,cir->ports);
  /*return S-parameters*/
  ytos(An->cans);

  /*If no jans present and jxy set, get current distribution*/
  if (An->jxy && An->jans==NULL)
    calc_jxy(result,cir,An,Box);

  if (stime)
  { stime = GetTime( stime );
    An->mat_solve_time = stime->user - start_time;
    if (An->State && An->mat_solve_time > 0)
      printf("    Matrix solve time = %d Seconds.\n",An->mat_solve_time);
  }

  Mat_close(result);

  Box->quad_prec = quad_prec_sto;

} /*An_circuit*/

qloads( Cs,Cf,Box,kcz_src,kcz_fld,S )  msub_list *Cs,*Cf;  Box_info *Box;
		  Vec_ptr kcz_src,kcz_fld;  Ssmat_ptr S;
/*************************************************************/
/* Load the system matrix.                                   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02oct88 */ 
/* jcr29mar89 modified for new Cir struct*/
{ int i,j;
  void hh_init();

  /*If different levels, do not do via to x,y coupling here*/
  for (i=0; i<NUM_ESUB; i++) for (j=i; j<NUM_ESUB; j++) if ( Cs==Cf || (i!=VIA)==(j!=VIA) )
  { if ( (*(Cs->mask+i) && *(Cf->mask+j)) || (*(Cf->mask+i) && *(Cs->mask+j)) )
    { init_do_coup(i,j,0,(wt_info **)NULL);
      qclearhh();
      hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
      qloadh( Box,cycles );
      qdo_coup(Cs,Cf,S);
    }
  }
  if ( Cs==Cf ) return;

  /*Look for Via source to x,y fld*/
  i=VIA;
  if ( *(Cs->mask+i) ) for (j=0; j<NUM_ESUB; j++) if ( (j!=VIA) && *(Cf->mask+j) )
  { init_do_coup(i,j,0,(wt_info **)NULL);
    qclearhh();
    hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
    qloadh( Box,cycles );
    qdo_coup(Cs,Cf,S);
  }

  /*Look for Via fld to x,y src*/
  j=VIA;
  if ( *(Cf->mask+j) ) for (i=0; i<NUM_ESUB; i++) if ( (i!=VIA) && *(Cs->mask+i) )
  { init_do_coup(i,j,0,(wt_info **)NULL);
    qclearhh();
    hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
    qloadh( Box,cycles );
    qdo_coup(Cs,Cf,S);
  }

  qclearhh();
} /*qloads*/

loads( Cs,Cf,Box,kcz_src,kcz_fld,S )  msub_list *Cs,*Cf;  Box_info *Box;
		  Vec_ptr kcz_src,kcz_fld;  Ssmat_ptr S;
/*************************************************************/
/* Load the system matrix.                                   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02oct88 */ 
/* jcr29mar89 modified for new Cir struct*/
{ int i,j;
  void hh_init();

  /*If different levels, do not do via to x,y coupling here*/
  for (i=0; i<NUM_ESUB; i++) for (j=i; j<NUM_ESUB; j++) if ( Cs==Cf || (i!=VIA)==(j!=VIA) )
  { if ( (*(Cs->mask+i) && *(Cf->mask+j)) || (*(Cf->mask+i) && *(Cs->mask+j)) )
    { init_do_coup(i,j,0,(wt_info **)NULL);
      clearhh();
      hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
      loadh( Box,cycles );
      do_coup(Cs,Cf,S);
    }
  }
  if ( Cs==Cf ) return;

  /*Look for Via source to x,y fld*/
  i=VIA;
  if ( *(Cs->mask+i) ) for (j=0; j<NUM_ESUB; j++) if ( (j!=VIA) && *(Cf->mask+j) )
  { init_do_coup(i,j,0,(wt_info **)NULL);
    clearhh();
    hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
    loadh( Box,cycles );
    do_coup(Cs,Cf,S);
  }

  /*Look for Via fld to x,y src*/
  j=VIA;
  if ( *(Cf->mask+j) ) for (i=0; i<NUM_ESUB; i++) if ( (i!=VIA) && *(Cs->mask+i) )
  { init_do_coup(i,j,0,(wt_info **)NULL);
    clearhh();
    hh_init(Box,Zhat,kcz_src,kcz_fld,Box->sub+i,Box->sub+j,cycles);
    loadh( Box,cycles );
    do_coup(Cs,Cf,S);
  }

  clearhh();
} /*loads*/

Rtyp min_lambda(Eps,Mu,Num,F)  Rtyp *Eps,*Mu,F;  int Num;
/*************************************************************/
/* Return the minimum open space wavelength of all layers.   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr09feb89 */ 
{ Rtyp prod;
  for ( prod=0.0; Num>0; Num--, Eps++, Mu++)
    if ( *Eps * *Mu > prod ) prod = *Eps * *Mu;
  return ( 1.0 / ( 1000.0 * F * sqrt(prod) ) );
} /*min_lambda*/

met_mod(cir,Box,F,S) Cir *cir;  Box_info *Box;  Rtyp F;  Ssmat_ptr S;
/*************************************************************/
/* Evaluate the metalization resistivity array.              */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21jun88 */ 
/* jcr31jan89 modified for generalized alg., incl. tri subs*/
{ int i,ist,jst,more;
  int ix0,iy0,jx0,jy0;
  int ix,iy,iloc,jloc;
  int isqr,jsqr,icond;
  int pos;
  Rtyp iwght,jwght;
  Rtyp *R,*X;  /*Resistance array*/
  Rtyp *Rp,*Xp; /*Pointer into Resistance array*/
  Met_type *Mp;  /*Pointer into Metalization array*/
  msub_list *lev;
  Poly_list *plev;
  msub *ip,*jp;
  sub_typ *isub,*jsub;
  Rtyp factor,via_mult;
  Rtyp tt_re,tt_im;
  int Re_part;

  if (Box->num_met<=0) return;

#ifdef EMPAR
  Re_part = 0;
  tt_re=0.0;
  tt_im=0.0;
  if (S->d && S->Re) Re_part = 1;
  if (!S->d && S->re) Re_part = 1;

#ifdef EM_MASPAR
  smat_mv_2_be(S);
#endif /* EM_MASPAR */
#endif /* EMPAR */

  /*Calculate each surface resistivity*/
  R = (Rtyp *)cnew( Box->num_met,sizeof(Rtyp) );
  X = (Rtyp *)cnew( Box->num_met,sizeof(Rtyp) );
  for (i=0; i<Box->num_met; i++)
  { Mp = Box->met + i;
    Rp = R + i;
    Xp = X + i;
    Do_res( Mp->rdc,Mp->rrf,F,Rp,Xp );
    *Xp += Mp->xdc;
  }

  /*Evaluate the resistivity constant for each sqr*/
  for (lev=cir->lev0,plev=cir->plev0; lev<cir->lev0+cir->n; lev++,plev++) if (lev->n>0)
  { via_mult = *(Box->H + (lev-cir->lev0)) * (Box->A/Box->Capm + Box->B/Box->Capn);
    for (ip=lev->msub0; ip<lev->msub0+lev->n; ip++)
      for (jp=lev->msub0; jp<lev->msub0+lev->n; jp++) if (ip->loc <= jp->loc && msub_overlapp(ip,jp))
	for (ist=0; ist<NUM_ESUB; ist++) for (isub=Box->sub+ist,jst=0; jst<NUM_ESUB; jst++) /*esub_typ loop*/
	{ jsub = Box->sub+jst;
          ix0 = -1;  iy0 = -1;
	  while( next_esub(ip,&ix0,&iy0,&iwght,isub) )
	  { jx0 = -1; jy0 = -1;
	    while( next_esub(jp,&jx0,&jy0,&jwght,jsub) )
              for (isqr=0; isqr<isub->num_sqr; isqr++)  /*sqr loop*/
                for (jsqr=0; jsqr<jsub->num_sqr; jsqr++)
	          /*calc sqr center coord*/
                  if ( (ix = ix0 + *(isub->sqr_coord+2*isqr))   ==
	               (     jx0 + *(jsub->sqr_coord+2*jsqr))   && 
                       (iy = iy0 + *(isub->sqr_coord+2*isqr+1)) ==
	               (     jy0 + *(jsub->sqr_coord+2*jsqr+1))     )
	          if ( ix>=0 && ix<=Box->Capm && iy>=0 && iy<=Box->Capn )
	          { iloc = *(isub->ej_loc+isqr);  jloc = *(jsub->ej_loc+jsqr);
	  	    factor = *(Box->ej_loss + iloc + jloc * Box->num_ej);
		    icond = cond_index_min(plev,ix,iy,R);
                    if (icond >= 0)
		    { /*edge mod for a single square, Area/3 -> Area/2 (uniform current)*/
		      if (ix==0 || iy==0 || ix==Box->Capm || iy==Box->Capn) factor *= 1.5;
		      factor *= iwght * jwght;
		      /*Do sym for all but center line x dir (2nd 2 condx) current*/
                      if (Box->Xsym && (iy0!=Box->Capn/2 || iy0!=iy)) factor *= 0.5;
                      if (ip->id==MVIA && jp==ip) factor *= via_mult;
                      pos = jp->loc + ((ip->loc*(2*(S->n) - ip->loc - 1))>>1);
#ifdef EMPAR
		      if (Re_part) tt_re = -factor* (*(R+icond));
		      tt_im = -factor* (*(X+icond));
#ifdef EM_MASPAR
		      Wsmat_put(S,pos,tt_re,tt_im);
#endif /* EM_MASPAR */
#ifdef CM5
	/*	      CM_put_dat( S,ip->loc,jp->loc,tt_re,tt_im );  */
		      (*CM_write_2_fe_buf)(S,ip->loc,jp->loc,tt_re,tt_im);
#endif /* CM5 */
#endif /* EMPAR */
#ifndef EMPAR
		      if (S->d)
                      { if (S->Re) *(S->Re+pos) -= factor * (*(R+icond));
                        *(S->Im+pos) -= factor * (*(X+icond));
                      }
                      else
                      { if (S->re) *(S->re+pos) -= factor * (*(R+icond));
                        *(S->im+pos) -= factor * (*(X+icond));
                      }
#endif /* (not) EMPAR */
                    }
                  }
          }
        }
  }
  sFree( (char *)R);  sFree( (char *)X);
} /*met_mod*/

Do_res( rdc,rrf,F,Rp,Xp ) Rtyp rdc,rrf,F;  Rtyp *Rp,*Xp;
/*************************************************************/
/* Evaluate the surface resistivity.                         */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01aug88 */
{ complex kzt,ans;

  rrf *= sqrt(1000000.0*F);

  /*No loss*/
  if ( fabs(rrf)<1.0e-10 && fabs(rdc)<1.0e-10 )
  { *Rp = *Xp = 0.0;  return; }

  /*Electromagnetically thick*/
  if ( fabs(rrf) > 50.0*fabs(rdc) )
  { *Rp = rrf;  *Xp = rrf;  return; }

  /*Electromagnetically thin*/
  if ( 5000.0*fabs(rrf) < fabs(rdc) )
  { *Rp = rdc;  *Xp = 0.0;  return; }

  /*Transistion region*/
  ans.Im = ans.Re = - rrf;
  kzt.Im = kzt.Re = - rrf / rdc;
  Cexp( &kzt, &kzt );
  kzt.Re -= 1.0;
  Cdiv( &ans, &kzt, &ans );
  *Rp = ans.Re;
  *Xp = ans.Im;
} /*Do_res*/

Mat_ptr merge_ports(R,ports)  Mat_ptr R;  port_list *ports;
/*************************************************************/
/* Merge port subsections together.                          */
/* (C) COPYRIGHT 1987 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr04sep87 */ 
/* jcr03oct88 - modified for new port description */
/* R has all port subsections in order in the upper left corner */
/* R has been reduced to just the port subsections */
{ Mat_ptr T;  /*Merged ports matrix*/
  port_typ *iport,*jport;
  int row,col,i,j,ilim,jlim;
  Rtyp cur_dir;  /*port current direction +/- 1.0*/

  /*Open the matrix for the merged data*/
  T = Mat_open(ports->n,ports->n);

  row=0; /*destination row*/
  iport = ports->port0;
  ilim = iport->nsub;
  for (i=0; i<R->Num_row; i++)
  { if (i>=ilim) { iport++;  row++;  ilim += iport->nsub; }
    jport = ports->port0;  jlim = jport->nsub;  col=0;  /*destination col*/
    for (j=0; j<R->Num_row; j++) /*look at only square portion of R*/
    { if (j>=jlim) { jport++;  col++;  jlim += jport->nsub; }
      cur_dir = ( (iport->iside & 1) ^ (jport->iside & 1) ) ? -1.0 : 1.0;
      /*Minus sign is from variational reaction integral for impedance*/
      Mat_put_re( row,col, -cur_dir*Mat_get_re(i,j,R) + Mat_get_re(row,col,T), T);
      Mat_put_im( row,col, -cur_dir*Mat_get_im(i,j,R) + Mat_get_im(row,col,T), T);
    }
  }
  return(T);
} /*merge_ports*/

calc_jxy(R,cir,An,Box) Mat_ptr R;  Cir *cir;  An_info *An;  Box_info *Box;
/*************************************************************/
/* Generate current distribution matrix.                     */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr09apr91 */ 
{ int row,i,j,ilim;
  msub_list *lev;
  msub *im;
  port_typ *iport;
  Rtyp cur_dir;  /*port current direction +/- 1.0*/
  Rtyp wght,dx,dy;

  dx = Box->A / Box->Capm;
  dy = Box->B / Box->Capn;

  /*Assumming An->jans == NULL*/
  An->jans = Mat_open(cir->ports->n,R->Num_col);

  /*Current is in Amps, convert to Amps/meter*/
  for (lev=cir->lev0; lev<cir->lev0+cir->n; lev++)
    for (im=lev->msub0; im<lev->msub0+lev->n; im++)
    { if (im->loc >= R->Num_col)
        announce_bug("calc_jxy(0)");
      wght = (Box->Xsym) ? 2.0 : 1.0;
      switch( im->id )  /*Set wght so we have total current in sub*/
      { case MXR :  wght *= 2.0; /* *(im->ip+3) - *(im->ip+1); <-- current/cell */
                    wght = 1.0/(dy*wght);
		    /*Subs crossing line of sym have only half current*/
		    if ( Box->Xsym && *(im->ip+3) > Box->Capn/2 )
		      wght *= (*(im->ip+3) - *(im->ip+1)) / (Box->Capn/2 - *(im->ip+1));
                    break;
        case MYR :  wght *= 2.0; /* *(im->ip+2) - *im->ip; <-- current/cell */
                    wght = 1.0/(dx*wght);
                    break;
        case MRFX :
        case MENDXP :
        case MENDXN :  wght *= 0.5 / dy;  break;
        case MRFY :
        case MENDYP :
        case MENDYN :  wght *= 0.5 / dx;  break;
      } 
      for (i=0; i<R->Num_row; i++)
        Mat_put(i,im->loc,wght*Mat_get_re(i,im->loc,R),wght*Mat_get_im(i,im->loc,R),R);
    }

  row=0; /*destination row*/
  iport = cir->ports->port0;
  ilim = iport->nsub;
  for (i=0; i<R->Num_row; i++)
  { if (i>=ilim)
    { iport++;  row++;  ilim += iport->nsub; }
    for (j=0; j<R->Num_col; j++)
    { cur_dir = ( iport->iside & 1 ) ? -1.0 : 1.0;
      Mat_put_re( row,j, cur_dir*Mat_get_re(i,j,R) + Mat_get_re(row,j,An->jans), An->jans);
      Mat_put_im( row,j, cur_dir*Mat_get_im(i,j,R) + Mat_get_im(row,j,An->jans), An->jans);
    }
  }

} /*calc_jxy*/
